# -*- coding: japanese-shift-jis-dos; tab-width: 4; -*-
#
# Wdic.pm - WDIC manager, to be used from wdic-fpw.
# $Id: Wdic.pm 1.3.3.9 2003/08/02 05:55:09 satomii Exp $
#
# DO NOT CHANGE THE ENCODING OF THIS FILE.
#
# Copyright (C) 2000-2003, rei@wdic.org.
# This program is distributed in the hope that it will be useful, but
# without any warranty. See the GNU General Public License for the details.
#
package Wdic;

require 5.005;
require Exporter;

use strict;
use IO::File;
use File::DosGlob 'glob';
use integer;

use vars qw(
	$VERSION @ISA @EXPORT @EXPORT_OK
	$TRACE $CHARSET $LOGFILEHANDLE
);

$VERSION = '1.3';
$TRACE = 1;			# 1 for status messages, 2 for debug messages.
$CHARSET = '';		# display charset for tracing.
$LOGFILEHANDLE;		# log file handle.

@ISA = qw(Exporter);
@EXPORT = qw(
	&Wdic::init_wdic($)
	&Wdic::uninit_wdic()
	&Wdic::sort_group_keys(@)
	&Wdic::get_group_list()
	&Wdic::get_group_info($)
	&Wdic::get_group_title($)
	&Wdic::get_file_title($)
	&Wdic::get_dir_name($)
	&Wdic::locate_word($;$$)
	&Wdic::define_word($)
);
@EXPORT_OK = qw(
	$TRACE
	$CHARSET
	$LOGFILEHANDLE
	&log(@)
);

use fields qw(
	directory
	group_order
	wdic_group
	wdic_dirs
	wdic_file
	word_list
	word_index
);

my $SJIS_ANY = '(?:[\x81-\x9f\xe0-\xfc][\x40-\x7e\x80-\xfc])';
my $SJIS_YOMI =
	'(?:(?:\x82[\x4f-\x58\x60-\x79\x82-\x9a])' .	# alnum.
	'|(?:\x82[\x9f-\xf1])' .						# hiragana.
	'|(?:\x83[\x40-\x7e\x80-\x96]))';				# katakana.

##------------------------------------------------------------------------
## STATIC METHODS

sub log(@)
{
	my @args;
	local $_;

	foreach (@_) {
		if (/[^\x00-\x7e]/) {
			$_ = Jcode::convert($_, $CHARSET);
		}
		push(@args, $_);
	}
	if ($LOGFILEHANDLE) {
		$LOGFILEHANDLE->print(@args);
		$LOGFILEHANDLE->flush();
	}
	print @args;
}

##------------------------------------------------------------------------
## CONSTRUCTOR AND DESTRUCTOR

sub new()
{
	my $proto = shift;
	my $class = ref($proto) || $proto;
	my $self = {
		directory => '',
		group_order => {},
		wdic_group => {},
		wdic_dirs => {},
		wdic_file => {},
		word_list => ()
	};
	bless($self, $class);
	$self->initialize($_[0]) if (@_);
	return($self);
}

sub DESTROY()
{
	# leave everything to GC...
}

##------------------------------------------------------------------------
## EXPORTED METHODS

#
# int init_wdic(char *directory)
# loads WDIC files from the specified directory and constructs the
# word list to be used later by other methods. returns the number of
# words indexed.
#
sub Wdic::init_wdic($)
{
	my $self = shift;
	my $dir = $_[0];
	my (@list, $path);

	$dir =~ s/^\\/\//g;
	$dir =~ s/\/$//;
	-d $dir || die "Not a valid direrctory: $dir ($!)";

	$self->uninit_wdic() if ('' ne $self->{directory});
	$self->{directory} = $dir;

	# load the group files.
	$self->init_groups();
	if (@list = glob("$dir/*.WLF")) {
		foreach $path (@list) {
			eval { $self->load_group($path) };
			warn $@ if ($@);
		}
	} else {
		warn "File not found: $dir/*.WLF\n";
	}
	if (1 < $TRACE) {
		&log("Dumping groups...\n");
		&dump_hash(0, 'wdic_group', %{ $self->{wdic_group} });
	}

	# load the directory listings if any.
	$path = glob("$dir/DIR.LST");
	if ($path || -f "$dir/DIR.LST") {
		$self->load_directory($path ? $path : "$dir/DIR.LST");
	}

	# load wdic files and construct word list.
	@list = glob("$dir/*.DIC");
	@list || die "File not found: $dir/*.DIC";
	foreach $path (@list) {
		eval { $self->load_file($path) };
		warn $@ if ($@);
	}

	# sort word entries.
	@list = @{ $self->{word_list} };
	@list || die "No word found in $dir/*.DIC.";

	&trace("Sorting " . scalar(@list) . " entries...");
	@list = sort {
		my %a = %{ $a };
		my %b = %{ $b };
		$a{SORT} cmp $b{SORT} || $a{WORD} cmp $b{WORD};
	} @list;
	$self->{word_list} = \@list;
	&trace(" done.\n");

	# make simple index for the word list.
	&trace("Making index...");
	$self->make_word_index();
	&trace(" done.\n");

	if (1 < $TRACE) {
		&trace("DEBUG: Wdic::initialize() is dumping class fields...\n");
		&trace('$directory = ' . $self->{directory} . "\n");
		&dump_hash(0, 'wdic_group', %{ $self->{wdic_group} });
		&dump_hash(0, 'wdic_file', %{ $self->{wdic_file} });
	}
	scalar(@{ $self->{word_list} });
}

#
# void uninit_wdic(void)
#
sub Wdic::uninit_wdic()
{
	my $self = shift;

	$self->{directory} = '';
	undef $self->{group_order};
	undef $self->{wdic_group};
	undef $self->{wdic_dirs};
	undef $self->{wdic_file};
	undef $self->{word_list};
}

#
# char **sort_group_keys(char **groups)
# sorts a list of group key names.
#
sub Wdic::sort_group_keys(@)
{
	my $self = shift;
	my %order = %{ $self->{group_order} };

	sort {
		if ($order{$a}) {
			$order{$b} ? $order{$a} <=> $order{$b} : 1;
		} else {
			$order{$b} ? -1 : $a cmp $b;
		}
	} @_;
}

#
# char **get_group_list(void)
# returns a list of group names currnetly loaded into this module.
#
sub Wdic::get_group_list()
{
	my $self = shift;
	my @list = $self->sort_group_keys(keys(%{ $self->{wdic_group} }));

	if (1 < $TRACE) {
		&dump_array(0, 'group_list', @list);
	}
	@list;
}

#
# (hash) get_group_info(char *group_name)
# returns a hash that contains the group information.
#
sub Wdic::get_group_info($)
{
	my $self = shift;
	my %group = %{ $self->{wdic_group} };
	defined($group{$_[0]}) ? %{ $group{$_[0]} } : ();
}

#
# char *get_group_title(char *group)
# returns the title of a group. if the group was not found in the group
# list, the method returns the same string as passed.
#
sub Wdic::get_group_title($)
{
	my $self = shift;
	my %group = %{ $self->{wdic_group} };
	my $name;

	$name = $group{$_[0]}{NAME} if (defined($group{$_[0]}));
	$name ? $name : $_[0];
}

#
# char *get_file_title(char *file)
# returns the title of a file (dictionary). if the file was not found
# in the file list, the method returns the same string as passed.
#
sub Wdic::get_file_title($)
{
	my $self = shift;
	my %file = %{ $self->{wdic_file} };
	my $name;

	$name = $file{$_[0]}{NAME} if (defined($file{$_[0]}));
	$name ? $name : $_[0];
}

#
# char **get_dir_list(char *parent, int all)
# returns a list of child directories under specified directory. if no
# argument is given, all directories are returned.
#
sub Wdic::get_dir_list($;$)
{
	my $self = shift;
	my ($regexp, $all) = @_;
	my @keys = keys(%{ $self->{wdic_dirs} });

	if ($regexp) {
		if ('/' eq $regexp) { $regexp = ''; }
		else { $regexp =~ s:/:\\/:g; }
		@keys = grep(/^$regexp\//, @keys);
	}

	unless ($all) {
		my $i = 0;
		my $len = 0;
		my $key = '';

		@keys = sort @keys;
		while ($i < scalar(@keys)) {
			if (0 < $len
					&& $len <= length($keys[$i])
					&& '/' eq substr($keys[$i], $len, 1)
					&& $key eq substr($keys[$i], 0, $len)) {
				splice(@keys, $i, 1);
			} else {
				$key = $keys[$i++];
				$len = length($key);
			}
		}
	}

	sort {
		$self->{wdic_dirs}{$a}{ID} <=> $self->{wdic_dirs}{$b}{ID};
	} @keys;
}

#
# (hash) get_dir_info(char *dir)
# returns a hash that contains the directory information.
#
sub Wdic::get_dir_info($)
{
	my $self = shift;
	my %dirs = %{ $self->{wdic_dirs} };
	defined($dirs{$_[0]}) ? %{ $dirs{$_[0]} } : ();
}

#
# char **locate_word(char *word, char *group, char *file)
# finds the specified word from the index and returns the group, file
# and offset in the array context.
#
sub Wdic::locate_word($;$$)
{
	my $self = shift;
	my ($word, $group, $file) = @_;
	my (@list, @found, $key);

	if ($group) {
		# if the specified group is not available, immediately return.
		return if (!defined($self->{wdic_group}{$group}));
	}

	@list = unpack('SS', $word);
	$key = sprintf("%04X", $list[0]);
	$key .= sprintf("%04X", $list[1]) if (defined($list[1]));

	return if (!$self->{word_index}{$key});
	@list = @{ $self->{word_index}{$key} };

	foreach $key (@list) {
		my %word = %{ $self->{word_list}[$key] };
		next if ($file && $file ne $word{FILE});
		next if ($group && $group ne $word{GROUP});
		if ($word eq $word{WORD}) {
			@found = ($key, $word{GROUP}, $word{FILE}, $word{OFFSET});
			last;
		}
	}
	@found;
}

#
# (hash) define_word(int index)
# returns one word definition at the specified index.
#
# %word = {
#     WORD => $word,
#     READ => $reading,
#     SPELL => $spell,
#     PRONOUNCE => $pronounce,
#     POS => $part_of_speech,
#     DEFINITION => @definitions,
#     COLUMN => @columns_or_examples,
#     REFERENCE => @references,
#     GROUP => $group,
#     DIRS => @dirs,
#     FILE => $file
# };
#
sub Wdic::define_word($)
{
	my $self = shift;
	my $index = $_[0];
	my $context = '';
	my (%word, %file, %define, @info, @def, @col, @ref, $info);
	local $_;

	$index < scalar(@{ $self->{word_list} })
		|| die "Invalid word index $index";
	%word = %{ ${@{ $self->{word_list} }}[$index] };
	%file = %{ ${%{ $self->{wdic_file} }}{$word{FILE}} };

	# seek to the word position and validate the word there.
	$file{HANDLE}->seek($word{OFFSET}, IO::File::SEEK_SET)
		|| die "$word{FILE}.DIC: Failed to seek to offset $word{OFFSET} ($!)";
	$_ = $file{HANDLE}->getline()
		|| die "$word{FILE}.DIC: File read error ($!)";
	if ($_ !~ s/^\#([^\r\n]+)[\r\n]+$/$1/ || $_ ne $word{WORD}) {
		die "$word{FILE}.DIC: $word{WORD}: Invalid offset $word{OFFSET}";
	}

	# read the next some lines.
	$info = '';
	while ($_ = $file{HANDLE}->getline()) {
		last if (!s/^ {8}([^\r\n]+)[\r\n]+$/$1/);
		if (/^E/) {
			push(@def, $_);
			last;
		} elsif (/^/) {
			push(@ref, $_);
			last;
		} else {
			$info .= $_;
		}
	}
	@info = &split_info($info);

	# read body lines of this word.
	while ($_ = $file{HANDLE}->getline()) {
		last if (!s/^ {8}([^\r\n]*)[\r\n]+$/$1/);

		#
		# since a column block may (unexpectedly) appear after the link
		# list, @col and/or @ref cannot be used to detect the current
		# context.
		#
		if (/^\x81[\x79\xa6]/) {
			# [y]
			push(@col, $_);
			$context = 'col';
		} elsif (/^/) {
			# ..., image block. this block ends by  and may contain a
			# plugin entry. currently, it is up to the caller how this
			# block is parsed.
			push(@col, $_);
			$context = 'col';
			while ($_ = $file{HANDLE}->getline()) {
				last if (!s/^ {8}([^\r\n]*)[\r\n]+$/$1/);
				push(@col, $_);
				last if (/^/);
			}
		} elsif (/^\x81[\xa8\x97\x9a\x9f\xa5\xa4]/) {
			# []
			push(@ref, $_);
			$context = 'ref';
		} elsif (/^E/) {
			# "E...", should be a description line.
			push(@def, $_);
			$context = '';
		} elsif ('ref' eq $context) {
			if (s/^  //) {
				$ref[scalar(@ref) - 1] .= $_;
			} else {
				push(@ref, $_);
			}
		} elsif ('col' eq $context) {
			push(@col, $_);
		} elsif (s/^  //) {
			# a description line, to be combined to the previous line.
			next if ('' eq $_);
			my $c = $def[$#def];
			while ($c =~ s/^($SJIS_ANY|[\x20-\x7e])//) {
				if ('' eq $c) {
					$c = $1;
					last;
				}
			}
			if (',' eq $c) {
				if (/^$SJIS_ANY/ || $def[$#def] =~ /$SJIS_ANY,$/) {
					$_ = " $_";
				}
			} elsif ('' ne $c
					&& $c !~ /^($SJIS_ANY|[^ \-\(\[])$/
					&& !/^($SJIS_ANY|[^ \-\)\]])/) {
				$_ = " $_";
			}
			$def[$#def] .= $_;
		} else {
			warn "$word{FILE}: Unexpected line: $_\n";
			if ('ref' eq $context) {
				push(@ref, $_);
			} else {
				push(@col, $_);
				$context = 'col';
			}
		}
	}

	return(
		WORD => $word{WORD},
		READ => $info[0] ? $info[0] : '',
		SPELL => $info[1] ? $info[1] : '',
		PRONOUNCE => $info[2] ? $info[2] : '',
		POS => $info[3] ? $info[3] : '',
		DEFINITION => \@def,
		COLUMN => \@col,
		REFERENCE => \@ref,
		GROUP => $word{GROUP},
		DIRS => $word{DIRS},
		FILE => $word{FILE},
		OFFSET => $word{OFFSET}
	);
}

##------------------------------------------------------------------------
## INDEX

sub Wdic::make_word_index()
{
	my $self = shift;
	my @list = @{ $self->{word_list} };
	my %index;

	for (my $i = 0; $i <= $#list; $i++) {
		my @key = unpack('SS', $list[$i]{WORD});
		my $key = sprintf("%04X", $key[0]);
		$key .= sprintf("%04X", $key[1]) if ($key[1]);
		push(@{ $index{$key} }, $i);
	}

	$self->{word_index} = \%index;
	1;
}

##------------------------------------------------------------------------
## LOADING GROUP FILE

sub Wdic::init_groups()
{
	my $self = shift;
	my $file = glob($self->{directory} . '/FILE.GL');
	my $handle = new
			IO::File($file ? $file : "$self->{directory}/FILE.GL", 'r');

	if ($handle) {
		my $count = 0;
		local $_;

		binmode $handle;
		while ($_ = $handle->getline()) {
			last if (/^\[GROUP\][\r\n]+$/);
		}
		while ($_ = $handle->getline()) {
			last if /^\[/;
			next if /^\s*\#/;
			s/[ \r\n]+$//;
			my ($key, $wlf, $name) = split(/\t+/, $_);
			if ($key) {
				$self->{group_order}{$key} = ++$count;
				$self->{wdic_group}{$key}{NAME} = $name ? $name : $key;
			}
		}
		$handle->close();

	} else {
		warn "Unable to open file: $file ($!)\n";
		$self->{group_order} = {
			WDIC => 1, TECH => 2, SCI => 3, MOE => 4,
			CUL => 5, MILI => 6, GEO => 7, RAIL => 8
		};
	}
}

sub Wdic::load_group($)
{
	my $self = shift;
	my (%group, $handle, $name, $section);
	local $_;

	$handle = new IO::File($_[0], 'r')
		|| die "Unable to open file: $_[0] ($!)";
	&trace("Reading $_[0]...\n");
	binmode $handle;

	$name = uc($_[0]);
	$name =~ s/^.*\/([^\/]+)\.WLF$/$1/;
	if (defined($self->{wdic_group}{$name})) {
		%group = %{ $self->{wdic_group}{$name} };
	}

	while ($_ = $handle->getline()) {
		s/[ \r\n]+$//;
		s/\s*\#.*$//;
		next if ('' eq $_);

		if (s/^\[(.+)\]$/$1/) {				# section declaration.
			$section = uc($_);
		} elsif ($section eq 'GROUP') {		# in [GROUP] section.
			if (s/^([^\=]+)\=//) {
				$group{uc($1)} = $_;
			}
		} elsif ($section eq 'FILES') {		# dictionary entry.
			my @fields = split(/\t/, $_);
			$fields[0] =~ s/\.DIC$//i;
			push(@{ $group{FILES} }, $fields[0]);
			$self->{wdic_file}{$fields[0]} = {
				GROUP => $name,
				NAME => $fields[1],
				CONTENT => $fields[2],
			};
		}
	}

	$handle->close();
	$self->{wdic_group}{$name} = \%group;
	1;
}

##------------------------------------------------------------------------
## LOADING DIRECTORY FILE

sub Wdic::load_directory($)
{
	my $self = shift;
	my $count = 0;
	my (%dirs, @list, $handle, $dir, $name);
	local $_;

	$handle = new IO::File($_[0], 'r')
		|| die "Unable to open file: $_[0] ($!)";
	&trace("Reading $_[0]...\n");
	binmode($handle);

	while ($_ = $handle->getline()) {
		s/[ \r\n]+$//;

		# DIR.LST in older WDIC may be encoded in JIS.
		$_ = Jcode::convert($_, 'sjis');

		# each line must be delimited by a tab character.
		@list = split(/\t+/, $_);
		$dir = shift(@list);
		$name = pop(@list);
		if (!$name) {
			$dir =~ /([^\/]+)$/;
			$name = $1;
		}

		# fill the hash.
		$dirs{$dir}{ID} = ++$count;
		$dirs{$dir}{NAME} = $name;
		$dirs{$dir}{REF} = $list[0] if (@list);
		$dirs{$dir}{LIST} = ();
	}

	$handle->close();
	$self->{wdic_dirs} = { %dirs };
	1;
}

##------------------------------------------------------------------------
## LOADING DICTIONARY FILE

sub Wdic::load_file($)
{
	my $self = shift;
	my (%entry, @list, @dirs, $file, $handle, $capital, $wc);
	local $_;

	$handle = new IO::File($_[0], 'r')
		|| die "Unable to open file: $_[0] ($!)";
	&trace("Indexing $_[0]...\n");
	binmode $handle;

	$file = uc($_[0]);
	$file =~ s/^.*\/([^\/]+)\.DIC$/$1/;
	%entry = %{ $self->{wdic_file}{$file} } if ($self->{wdic_file}{$file});

	if (!%entry) {
		%entry = (
			GROUP => '',
			NAME => '',
			CONTENT => '',
			CONTACT => '',
			EMAIL => ''
		);
		warn "$file not found in any group\n";
	}
	if (!$entry{GROUP} || !$entry{NAME}) {
		# determine the dictionary name.
		my $group;
		while ($_ = $handle->getline()) {
			last if (/^[\<\-%]/);
			if (/^\s*($SJIS_ANY+(?:̊bm)?)\s*(?:\x82[\x4f-\x58])+\s*([^\s]+).*[\r\n]+$/) {
				$group = $1;
				$entry{NAME} = $2 if (!$entry{NAME});
				last;
			}
		}
		if (!$entry{GROUP}) {
			foreach (keys(%{ $self->{wdic_group} })) {
				if ($group eq $self->{wdic_group}{$_}{NAME}) {
					$entry{GROUP} = $_;
					push(@{ $self->{wdic_group}{$_}{FILES} }, $file);
					last;
				}
			}
			if (!$entry{GROUP}) {
				$_ = $file;
				$entry{GROUP} = $file;
				if (!$self->{wdic_group}{$file}
						|| !$self->{wdic_group}{$file}{NAME}) {
					$self->{wdic_group}{$file}{NAME} = $group;
				}
				push(@{ $self->{wdic_group}{$file}{FILES} }, $file);
			}
		}
	}

	# construct the word list.
	$capital = '';
	$wc = 0;

	while ($_ = $handle->getline()) {
		if (s/^%([^\r\n]+)[\r\n]*$/$1/) {			# %...
			if ($_ eq 'L') {
				$capital = '0000';
			} elsif ($_ eq '') {
				$capital = '0001';
			} else {
				$capital = sprintf("%02x%02x", unpack('CC', $_));
			}
			@dirs = ();

		} elsif (s/^(\/[^\r\n]+)[\r\n]*$/$1/) {		# direrctory
			push(@dirs, $_);

		} elsif (/^#([^\r\n]+)[\r\n]*$/) {			# #...
			my %word = (
				WORD => $1,
				SORT => '',
				GROUP => $entry{GROUP},
				DIRS => [ @dirs ],
				FILE => $file,
				OFFSET => $handle->tell() - length($_)
			);

			(0 < $word{OFFSET}) || die "Read error: $_[0] ($!)\n";
			my ($info, $read);

			foreach (@dirs) {
				if ($self->{wdic_dirs}{$_}) {
					push(@{ $self->{wdic_dirs}{$_}{LIST} }, $word{WORD});
				}
			}

			# read information line(s).
			$_ = $handle->getline();
			if (!s/^ {8}([^\r\n]+)[\r\n]+$/$1/) {
				warn "Information line not found: $word{WORD}\n";
				next;
			}

			$info = $_;
			while ($info !~ /l$/) {
				$_ = $handle->getline();
				last if (!$_
					|| !s/^ {8}([^ ][^\r\n]*)[\r\n]+$/$1/
					|| /^\x81[\x45\xa8]/);		# /^[E]/
				$info .= $_;
			}

			# store the information into a hash.
			$read = ($info =~ /^\[((?:$SJIS_ANY|[^\]])+)\]/) ? $1 : '';
			if ($capital =~ /^00/ || $capital =~ /^82[67]/) {
				$word{SORT} = $capital . &make_sort_word($word{WORD});
			} elsif (!$read || $read =~ /^\?+$/) {
				$word{SORT} = &make_sort_word($word{WORD});
			} else {
				$word{SORT} = &make_sort_word($read);
			}

			# push this element to the list.
			push(@{ $self->{word_list} }, \%word);
			$wc++;
			@dirs = ();
		}
	}

	if (!$wc) {
		$handle->close();
		die "No word found in the file '$_[0]'";
	}

	# update the class field.
	$entry{HANDLE} = $handle;
	$self->{wdic_file}{$file} = \%entry;
	$wc;
}

##------------------------------------------------------------------------
## LOCAL FUNCTIONS

#
# char *make_sort_word(char *reading)
# converts Shift JIS characters, hiragana to katakana, lower to upper
# for case-insensitive sorting.
#
sub make_sort_word($)
{
	my @chars = unpack('C*', $_[0]);
	my $str = '';
	my ($c, $c2);

	while ($c = shift(@chars)) {
		if (0x81 == $c) {
			$c2 = shift(@chars);
			if (0x40 == $c2 || 0x45 == $c2 || 0x5b == $c2 || 0x7c == $c2) {
				# [@E[|], nop.
			} else {
				$str .= pack('CC', $c, $c2);
			}
		} elsif (0x82 == $c) {
			$c2 = shift(@chars);
			if (((0x9f <= $c2 && $c2 <= 0xa7) && ($c2 % 2))
					|| 0xc1 == $c2
					|| 0xe1 == $c2 || 0xe3 == $c2 || 0xe5 == $c2
					|| 0xec == $c2) {
				# []
				$c2++;
			} elsif (((0xaa <= $c2 && $c2 <= 0xc0) && !($c2 % 2))
					|| 0xc3 == $c2 || 0xc5 == $c2 || 0xc7 == $c2
					|| ((0xce <= $c2 && $c2 <= 0xda) && (2 == $c2 % 3))) {
				# [Âłǂ΂тԂׂ]
				$c2--;
			} elsif ((0xcf <= $c2 && $c2 <= 0xdb) && !($c2 % 3)) {
				# [ς҂Ղ؂]
				$c2 -= 2;
			}
			if (0x81 <= $c2 && $c2 <= 0x9a) {
				$c2 -= 0x21;
			} elsif (0x9f <= $c2 && $c2 <= 0xdd) {
				$c = 0x83;
				$c2 -= 0x5f;
			} elsif (0xde <= $c2 && $c2 <= 0xf1) {
				$c = 0x83;
				$c2 -= 0x5e;
			}
			$str .= pack('CC', $c, $c2);
		} elsif (0x83 == $c) {
			$c2 = shift(@chars);
			if (((0x40 <= $c2 && $c2 <= 0x48) && !($c2 % 2))
					|| 0x62 == $c2
					|| 0x83 == $c2 || 0x85 == $c2 || 0x87 == $c2
					|| 0x8e == $c2) {
				# [@BDFHb]
				$c2++;
			} elsif (0x95 == $c2) {
				# 
				$c2 = 0x4a;
			} elsif (0x96 == $c2) {
				# 
				$c2 = 0x50;
			} elsif (((0x4b <= $c2 && $c2 <= 0x61) && ($c2 % 2))
					|| 0x64 == $c2 || 0x66 == $c2 || 0x68 == $c2
					|| ((0x6f <= $c2 && $c2 <= 0x7b) && !($c2 % 3))) {
				# [KMOQSUWY[]_adfhorux{]
				$c2--;
			} elsif ((0x70 <= $c2 && $c2 <= 0x7c) && (1 == $c2 % 3)) {
				# [psvy|]
				$c2 -= 2;
			}
			$str .= pack('CC', $c, $c2);
		} elsif ((0x84 <= $c && $c <= 0x9f) || (0xe0 <= $c && $c <= 0xfc)) {
			$c2 = shift(@chars);
			$str .= pack('CC', $c, $c2);
		} elsif (0x20 == $c || 0x2d == $c) {
			# [ -], nop.
		} else {
			$str .= pack('C', $c);
		}
	}
	$str;
}

#
# char **split_info(char *info)
# extracts reading, english, pronounce and POS from the word information
# line.
#
sub split_info($)
{
	my $str = $_[0];
	my @info = ('', '', '', '');
	my (@chars, $paren, $c, $i);

	if ($str =~ /^\[((?:$SJIS_ANY|[^\]])+)\]/) {
		$info[0] = $1;
	}
	if ($str =~ s/k((?:$SJIS_ANY|[@\+\/])+)l//) {
		#
		#k...l, the part of speech. '/' is not a valid char here, but
		# contained in the RC5 version of APPL.DIC..
		#
		$info[3] = $1;
	}

	@chars = unpack('C*', $str);
	$str = '';
	$paren = 0;

	for ($i = 0; $i < scalar(@chars); $i++) {
		$c = $chars[$i];
		if ((0x81 <= $c && $c <= 0x9f) || (0xe0 <= $c && $c <= 0xfc)) {
			$i++;
			if ($paren) {
				$info[1] .= pack('CC', $c, $chars[$i]);
			} elsif ('/' eq $str) {
				$info[2].= pack('CC', $c, $chars[$i]);
			}
			next;
		}
		if ('/' eq $str) {
			last if (0x2f == $c);		# /
			$info[2] .= pack('C', $c);
		} elsif (!$paren && '' eq $str && 0x2f == $c) {
			$str = '/';
		} else {
			if (0x28 == $c) {			# (
				$paren++;
				next if (1 == $paren);
			} elsif (0x29 == $c) {		# )
				$paren--;
				next if (!$paren);
			}
			$info[1] .= pack('C', $c) if (0 < $paren);
		}
	}

	$info[2] =~ s/ *[,\|] +\- +/\|-/g;
	$info[2] =~ s/ *[,\|] +/\|/g;
	$info[2] =~ s/ +(\[$SJIS_ANY\])/$1/g;
	@info;
}

##------------------------------------------------------------------------
## DEBUG FUNCTIONS

sub trace($)
{
	my $str = shift;

	return if (!$TRACE);
	$str = sprintf($str, @_) if (@_);
	if ($str =~ /[^\x00-\x7e]/ && '' ne $CHARSET && $INC{'Jcode.pm'}) {
		$str = Jcode::convert($str, $CHARSET);
	}
	&log($str);
}

sub dump_array($$@)
{
	my ($indent, $name, @array) = @_;
	my ($i, $line);

	&log(' ' x $indent . "\@$name =");
	if (!@array) {
		&log(" (empty)\n");
		return;
	}
	&log("\n");

	for ($i = 0; $i < scalar(@array); $i++) {
		if (!defined($array[$i])) {
			$line = '(undef)';
		} elsif (ref($array[$i]) =~ "HASH") {
			&dump_hash($indent + 2, "[$i]", %{ $array[$i] });
			next;
		} elsif (ref($array[$i]) eq "ARRAY") {
			&dump_array($indent + 2, "[$i]", @{ $array[$i] });
			next;
		} else {
			$line = $array[$i];
		}
		if ($line =~ /[^\x00-\x7e]/
				&& '' ne $CHARSET
				&& defined($INC{'Jcode.pm'})) {
			$line = Jcode::convert($line, $CHARSET);
		}
		&log(' ' x ($indent + 2) . "[$i] = $line\n");
	}
}

sub dump_hash($$%)
{
	my ($indent, $name, %hash) = @_;
	my ($key, $value, $line);

	&log(' ' x $indent . "\%$name =");
	if (!%hash) {
		&log(" (empty)\n");
		return;
	}
	&log("\n");

	while (($key, $value) = each(%hash)) {
		if (!defined($value)) {
			$line = '(undef)';
		} elsif (ref($value) =~ "HASH") {
			&dump_hash($indent + 2, $key, %{ $value });
			next;
		} elsif (ref($value) eq "ARRAY") {
			&dump_array($indent + 2, $key, @{ $value });
			next;
		} else {
			$line = $value;
		}
		if ($line =~ /[^\x00-\x7e]/
				&& '' ne $CHARSET
				&& defined($INC{'Jcode.pm'})) {
			$line = Jcode::convert($line, $CHARSET);
		}
		&log(' ' x ($indent + 2) . "$key => $line\n");
	}
}

1;
